#include <omp.h>
#include <stdio.h>
#include <cuda_runtime.h>
#include <unistd.h>

#include "util.h"
#include "log.h"
#include "arguments.h"
#include "GPGPU.h"
#include "arrayOps.h"


/* The kernel
 * Each GPU thread will run this kernel
 * @param A the first array of input data
 * @param B the second array of input data
 * @param C the array to store the results
 * @param numElements the number of elements in the arrays
 */
__global__ void vectorAdd ( float *A, float *B, float *C, int numElements )
{
  int index;

  // TODO 6.6 Calculate the index
  // Each thread will perform one addition, but which element of the arrays should this thread do?
  // Hint: you can calculate the index of the thread using blockIdx and blockDim.

  // TODO 6.7 Calculate do the addition and store the result in C
}

/*
 * Main - the entry point of the program
 */
int main ( int argc, char **argv )
{
  //The number of elements in each array
  unsigned int numEls = 10000000;

  // The number of threads to use
  int noThreads = 1;

  // Beginning and end of multiple test
  unsigned int begin  = 0;
  unsigned int end    = 0;
  unsigned int step   = 1;
  float noTest        = 20;

  // Initialise some command line arguments
  commandLineArg args[] =
  {
    // TYPE         char  Text        type            Explanation                                                 value
    { ARG_HELPTEXT, 0,    "",         "",             "Usage: ",                                                  NULL },
    { ARG_HELPTEXT, 0,    "",         "",             "A toy problem to demonstrate CUDA vector addition",        NULL },
    { ARG_NONE,     'h',  "help",     "",             "Display the help message",                                 NULL },
    { ARG_NONE,     'l',  "log",      "",             "Output a csv log file",                                    NULL },
    { ARG_HELPTEXT, 0,    "",         "",             "",                                                         NULL },
    { ARG_UINT,     'n',  "numEls",   "Number",       "Specify the number of elements",                           (void*)&numEls    },
    { ARG_INT,      't',  "threads",  "Number",       "Specify the number of CPU thread to use",                  (void*)&noThreads },
    { ARG_HELPTEXT, 0,    "",         "",             "\nMultiple problem sizes",                                 NULL },
    { ARG_UINT,     'b',  "begin",    "Number",       "Run 20 test beginning at NUMBER, you must specify end",    (void*)&begin     },
    { ARG_UINT,     'e',  "end",      "Number",       "Run 20 test ending at NUMBER.",                            (void*)&end       },
    { ARG_END,      0,    "",         "",             "",                                                         NULL }

  };
  char *name = strrchr ( argv[0], 47 );                           // find slash
  name = name != NULL ? ++name : name = argv[0];                  // get the name of the command run
  sprintf ( args[0].description, "Usage: %s [OPTION]...", name ); // Set the name of the program for the usage message

  // Get the host name of commuter the program is run on (for node identification)
  char hostname[1024];
  gethostname ( hostname, 1024 );

  // A command Line Argument: name, type and value
  commandLineArg readArg;

  // Set up cvs log file
  Logger *logger = new Logger();

  // The size in bytes of each array
  unsigned int noBytes;

  /// Time to perform various tasks [ms]
  double CPUcalTime, GPUcalTime, CPUmallocTime, GPUmallocTime, GPUcopyTime, copyTimeD2H;

  // Error value to evaluate CUDA calls
  cudaError_t result;

  // Timer used for timing how long various tasks take
  timer timer;

  // Menu, read any command line arguments and handle them
  {
    int argN = 1;         // Starts with the first argument 1 ( 0 is the name if the binary)

    while ( argN < argc ) // Loop through arguments
    {
      // read the next command line argument (note this increments argN)
      readArg = readNextArg ( &argN, argc, argv, args );

      if ( readArg.type != ARG_END )
      {
        // Argument was found in the list args, so handle it
        if ( readArg.charName == 'h' )
        {
          printHelpFromArgs ( args );         // Print help message
          exit ( EXIT_SUCCESS );              // Exit
        }
        else if ( readArg.charName == 'l' )
        {
          // get the time and create the log file
          time_t rawtime;
          tm* ptm;
          time ( &rawtime );
          ptm = localtime ( &rawtime );
          char timeMsg[1024];
          sprintf ( timeMsg, "runLog_%04i-%02i-%02i_%02i_%02i_%02i.csv", 1900 + ptm->tm_year, ptm->tm_mon + 1, ptm->tm_mday, ptm->tm_hour, ptm->tm_min, ptm->tm_sec );
          logger->setFile ( timeMsg ); // Once this is done the write commands will write to the file.
        }
      }
      else
      {
        // Invalid parameter
        printHelpFromArgs ( args );         // Print help message
        exit ( EXIT_FAILURE );              // Exit
      }
    }
  }

  printf ( "\n  ---=== CUDA Vector Add ===--- \n\n" );

  // Host memory pointers
  float* hArrayA = NULL;
  float* hArrayB = NULL;
  float* hArrayC = NULL;
  float* hArrayD = NULL;

  // Device memory pointers
  float* dArrayA = NULL;
  float* dArrayB = NULL;
  float* dArrayC = NULL;

  // If we are doing multiple steps decide on the step size and start and end parameters
  if ( begin > 0 || end > 0 )
  {
    if ( begin ==  0 )
    {
      step = ceil ( end / ( noTest + 1 ) );
      begin = step;
    }
    else if ( end == 0 )
    {
      fprintf ( stderr, "ERROR: You are must specify and end if specifying a start.\n" );
      exit ( EXIT_FAILURE );
    }

    if ( begin >= end )
    {
      fprintf ( stderr, "ERROR: You end must be grate than start.\n" );
      exit ( EXIT_FAILURE );
    }

    step = ceil ( ( end - begin ) / noTest );
    end = begin + step * noTest;
  }

  // TODO 6.11 Make this block run multiple times
  // This can easily be done by using the -e flag
  for ( int i = begin; i <= end; i += step )
  {
    if ( i != 0 )
    {
      numEls = i;
    }

    // Calculate the size in bytes of a single arrays
    noBytes = numEls * sizeof ( float );

    printf ( "\n Summing two vectors of %.1fM elements\n", numEls * 1e-6 );
    printf ( "------------------------------------------------------------------------------\n" );

    printf ( "[CPU]\n" );
    logger->csvWrite ( "host;numEls;noBytes;Gb", ";#;bytes;Gb", "%s;%d;%d;%0.3f", hostname, numEls, noBytes, noBytes * 1e-9 );

    // Check for free memory and allocate arrays
    float noGb = noBytes * 1e-9 * 4;

    if ( noBytes * 4 > getFreeRam() * 0.95 )
    {
      // Trying to allocate too much memory, this is a toy problem after all!
      fprintf ( stderr, "ERROR: You are trying to allocate %0.1fGb of host Memory and there is only %0.1fGb available, this may cause problems!\n", noGb, getFreeRam() * 1e-9 );
      exit ( EXIT_FAILURE );
    }
    else
    {
      printf ( "Allocating %2.1fGb (%4.1f%%) of host Memory (4 x %0.1fGb)     ...", noGb, ( float ) noBytes * 4.0f / ( float ) getTotalRam() * 100.0f, noBytes * 1e-9 );
      std::cout.flush();

      timer.Start();                // Start a time

      // Allocate the host memory
      hArrayA = ( float* ) malloc ( noBytes );
      hArrayB = ( float* ) malloc ( noBytes );
      hArrayC = ( float* ) malloc ( noBytes );
      hArrayD = ( float* ) malloc ( noBytes );

      CPUmallocTime = timer.stopS();  // Stop the timer
      printf ( " Complete in %.3fms\n", CPUmallocTime * 1e3 );
      logger->csvWrite ( "host alloc", "s", "%.6f", CPUmallocTime );
    }

    // Fill host array A and B with random numbers
    {
      timer.Start();                // Start a time
      printf ( "Generating %5.1fM elements in array A                   ...", numEls * 1e-6 );
      std::cout.flush();
      randArray ( hArrayA, numEls, 0.0f, 1.0f );
      printf ( " Complete in %0.3fms.\n", timer.stopS() * 1e3 );
      logger->csvWrite ( "Gen 01", "s", "%.6f", timer.stopS() );

      timer.Start();                // Start a time
      printf ( "Generating %5.1fM elements in array B                   ...", numEls * 1e-6 );
      std::cout.flush();
      randArray ( hArrayB, numEls, 0.0f, 1.0f );
      printf ( " Complete in %.3fms.\n", timer.stopS() * 1e3 );
      logger->csvWrite ( "Gen 02", "s", "%.6f", timer.stopS() );
    }

    // Vector Addition on the CPU    C = A + B
    {
      printf ( "Adding vectors of %5.1fM elements each                  ...", ( float ) ( numEls * 1e-6 ) );
      std::cout.flush();

      int holdThreads = 0;

      if ( noThreads > 0 )
      {
        holdThreads = omp_get_max_threads();
        omp_set_num_threads ( noThreads );
      }

      timer.Start();                 // Start a time

      // The actual number of threads used to do the addition on the CPU
      int noThreadsUsed = -1;

      //#pragma omp parallel                                  // This lines makes the below for loop run in parallel on the CPU
      {
        // Master thread sets noThreadsUsed
        if ( omp_get_thread_num() == 0 )
          noThreadsUsed = omp_get_num_threads();

        // Do the actual summing with a for loop
        #pragma omp for
        for ( unsigned long long i = 0; i < numEls; i++ )
        {
          hArrayC[i] = hArrayA[i] + hArrayB[i];             // C = A + B
        }
      }

      CPUcalTime = timer.stopS();      // Stop the time
      printf ( " Complete in %.3fms\n", CPUcalTime * 1e3 );
      printf ( "Used %d thread(s) on %d CPU processor(s)\n", noThreadsUsed, omp_get_num_procs() );
      logger->csvWrite ( "numProcs;numThreads;CPU add", "#;#;s", "%d;%d;%.6f", omp_get_num_procs(), noThreadsUsed, CPUcalTime );

      if ( holdThreads > 0 )
        omp_set_num_threads ( holdThreads );
    }

    // That concludes the vector addition on the CPU now for the GPU...

    printf ( "\n[GPU]\n" );

    // set some timing variables to 0
    GPUcopyTime = 0;
    GPUcalTime = 0;
    copyTimeD2H = 0;

    // TODO 6.1 Specify which CUDA device to use
    // This will be on the paper with you're user name and passwords.
    // The resource manager can not manage GPU's so we have to specify the device to negate contention for GPU's
    {
      TODOMSG("Specify the correct device to use")
    }

    timer.Start();

    // TODO 6.2 Allocate device memory for vectors A B and C    (dArrayA, dArrayB, dArrayC)
    {
      // You should allocate device memory and assign: dArrayA, dArrayB and dArrayC for later use.
      TODOMSG("Allocate device memory")
    }
    GPUmallocTime = timer.stopS();      // Stop the time
    logger->csvWrite ( "GPU malloc", "s", "%.6f", GPUmallocTime );

    // Copy input data from the host to the device
    {
      printf ( "Copying %0.2fGb to device                                ...", noBytes * 2 * 1e-9 );
      std::cout.flush();
      timer.Start();

      // TODO 6.3 Copy input data from the host to the device         (dArrayA, dArrayB)
      {
        // Copy values from hArrayA and hArrayB on to the GPU using dArrayA, dArrayB.
        TODOMSG("Copy input data from the host to the device")
      }

      // Some output
      {
        result = cudaDeviceSynchronize();   // Wait for the device to finish copying memory
        GPUcopyTime += timer.stopS();       // Stop the time

        if ( result != cudaSuccess )
          fprintf ( stderr, "\nERROR: Error copying data to device [%s]\n", cudaGetErrorString ( result ) );
        else
          printf ( " Complete in %.3fms\n", GPUcopyTime * 1e3 );
      }
    }

    dim3 dimBlock, dimGrid;
    int threadsPerBlock, blocksPerGrid;

    // Initialise grid
    {
      // TODO 6.4 setup the blocks and grid
      {
        // Set up blocks
        // Here you will have to set up the variables: dimBlock and threadsPerBlock
        TODOMSG("Calculate block dimensions")

        // Set up grid
        // Here you will have to set up the variables: dimGrid and blocksPerGrid
        TODOMSG("Calculate grid dimensions")
      }
    }

    // Call the kernel
    {
      timer.Start();

      // TODO 6.5 Call the kernel
      {
        // Call the kernel, you should use dimGrid and dimBlock, set up in the previous sections
        // and pass in the address of the device memory
        TODOMSG("Launch the kernel")
      }

      // Run message
      {
        result = cudaGetLastError();  // This determines whether the kernel was launched

        if ( result == cudaSuccess )
          printf ( "Running kernel                                          ..." );
        else
        {
          fprintf ( stderr, "ERROR: Error at kernel launch %s\n", cudaGetErrorString ( result ) );
          exit ( EXIT_FAILURE );
        }
      }

      // Wait for the device to finish running the code
      {
        result = cudaDeviceSynchronize();  // This will return when the kernel computation is complete, remember asynchronous execution

        // Complete message
        if ( result == cudaSuccess )
        {
          GPUcalTime += timer.stopS();
          printf ( " Complete in %.3fms\n", GPUcalTime * 1e3 );
        }
        else
          fprintf ( stderr, "\nERROR: Error after kernel launch %s\n", cudaGetErrorString ( result ) );
      }
    }

    // Copy result to host
    {
      printf ( "Copying %.2fGb from device                              ...", noBytes * 1e-9 );
      std::cout.flush();

      timer.Start();                      // Start the time

      // TODO 6.8 Copy device memory to host
      {
        // The kernel should have done the calculations and stored the result on device memory, now you need to get this back
        // to host memory so you can work with it.
        TODOMSG("Copy device memory to host")
      }
      // Wait for the device to finish copying memory
      {
        result = cudaDeviceSynchronize(); // This will return when the copy is complete
        copyTimeD2H += timer.stopS();      // Stop the timer
        GPUcopyTime += copyTimeD2H;

        if ( result != cudaSuccess )
          fprintf ( stderr, "\nERROR: Error copying data to device [%s]\n", cudaGetErrorString ( result ) );
        else
          printf ( " Complete in %.3fms\n", copyTimeD2H * 1e3 );
      }

      // Log some output
      logger->csvWrite ( "H->D;GPU add;D->H;GPU TIME", "s;s;s;s", "%.6f;%.6f;%.6f;%.6f", GPUcopyTime, GPUcalTime, copyTimeD2H, GPUcopyTime + GPUcalTime + copyTimeD2H );
    }

    // Validate - calculate if the arrays differ and output some stats
    {
      float differnce;
      bool same = true;

      printf ( "Comparing results                                       ..." );
      std::cout.flush();

      timer.Start();

      #pragma omp parallel for private(differnce)         // Do the check in parallel on the CPU
      for ( int i = 0; i < numEls; ++i )
      {
        differnce = fabs ( hArrayC[i] - hArrayD[i] );

        if ( differnce > 1e-7 )
        {
          //fprintf ( stderr, "ERROR: Difference %f at %06d is to great verification failure! \n", difference, i );
          #pragma omp automic
          same = false;
        }
      }

      printf ( " Complete in %.3fms\n", timer.stopS() * 1e3 );
      logger->csvWrite ( "compare", "s", "%.6f", timer.stopS() );

      if ( same )
      {
        // Print out some stats
        fprintf ( stdout, "Verification Passed \n" );

        float GPUflops = fDivide ( numEls, GPUcalTime );  // one flop (+) per element
        float CPUflops = fDivide ( numEls, CPUcalTime );  // one flop (+) per element

        float compSpeedup = fDivide ( CPUcalTime, GPUcalTime );
        printf ( "Computation speed-up %.2f", compSpeedup );

        float actualSpeedup = fDivide ( CPUcalTime, ( GPUcalTime + GPUcopyTime ) );
        printf ( " real speed-up %.2f \n", actualSpeedup );
        printf ( "CPU Gflops %.1f   GPU Gflops %.1f\n", CPUflops * 1e-9, GPUflops * 1e-9 );

        logger->csvWrite ( "compu speed;real speedup", "x;x", "%.4f;%.4f", compSpeedup, actualSpeedup );
        logger->csvWrite ( "GPU Gflops;CPU Gflops", "Gflops;Gflops", "%.2f;%.2f", GPUflops * 1e-9, CPUflops * 1e-9 );
        logger->csvWrite ( "veri", "bool", "1" );
      }
      else
      {
        fprintf ( stdout, "Verification Failed \n" );
        logger->csvWrite ( "compu speed;real speedup", "x;x", -1, -1 );
        logger->csvWrite ( "veri", "bool", "0" );
      }
    }

    // Free memory's
    {
      //TODO 6.9 free device memory's!
      {
        // Free all the device you allocated
        TODOMSG("Free device memory")
      }

      // Free host memory
      free ( hArrayA );
      free ( hArrayB );
      free ( hArrayC );
      free ( hArrayD );
    }
    printf ( "------------------------------------------------------------------------------\n" );

    logger->csvEndLine(); // End an iteration with this to create a table in the cvs log file
  }

  CUDA_SAFE_CALL ( cudaDeviceReset(), "Error resetting device" );

  printf ( "Complete\n" );

  delete ( logger );

  exit ( EXIT_SUCCESS );
}
