#ifndef ARRAYOPS_H_INCLUDED
#define ARRAYOPS_H_INCLUDED

#include <omp.h>
#include <stdio.h>
#include <unistd.h>
#include <math.h>

#if _WIN32
#include <Windows.h>
#elif __linux
#include <sys/time.h>
#endif

#include "utilstats.h"

/** A function to fill an array with random numbers, this uses 
 *
 * @param array   A pointer to the array to be filled, this should allocated
 * @param noEls   the number of element to generate
 * @param min     The minimum value of the randomly generate numbers
 * @param max     The maximum value of the randomly generate numbers
 * @return void
 **/
void randArray   ( float* array, const unsigned long noEls, const float min, const float max );

/** A function to fill an array with random numbers, this uses 
 *
 * @param array   A pointer to the array to be filled, this should allocated
 * @param noEls   the number of element to generate
 * @param min     The minimum value of the randomly generate numbers
 * @param max     The maximum value of the randomly generate numbers
 * @return void
 **/
void randArray ( double* array, unsigned long noEls, double min, double max );

/** A function to fill an array with random numbers, this uses 
 *
 * @param array   A pointer to the array to be filled, this should allocated
 * @param noEls   the number of element to generate
 * @param min     The minimum value of the randomly generate numbers
 * @param max     The maximum value of the randomly generate numbers
 * @return void
 **/
void randArray ( int* array, unsigned long noEls, int min, int max );

/* A function to print (to std out) array multiplication
 * @param array a pointer to the array to be filled, this should allocated
 * @param noEls the number of element to generate
 * @param min The minimum value of the randomly generate numbers
 * @param max The maximum value of the randomly generate numbers
 */
void printArray ( int *arrayA, int *arrayB, int *arrayC, unsigned int m, unsigned int n, unsigned int o );

/* A function to print (to std out) an array multiplication
 * @param arrayA First array m x n
 * @param arrayB Second array n x o
 * @param arrayB The Product array m x o
 */
void printArray ( float *arrayA, float *arrayB, float *arrayC, unsigned int m, unsigned int n, unsigned int o );

/** Calculate the sum of all absolute differences between all corresponding entries
 * This is calculated on the CPU in parallel
 * \f$ \sum_{i=0}^{n-1} {|A_i-B_i|}  \f$
 * 
 * @param arrayA The first array
 * @param arrayB The second array
 * @param n The number of elements to compare
 */
float absDiff ( float *arrayA, float *arrayB, unsigned long n );

/** Calculate the maximum absolute differences between ant two corresponding entries
 * This is calculated on the CPU in parallel
 * 
 * @param arrayA The first array
 * @param arrayB The second array
 * @param n The number of elements to compare
 */
float maxAbsDiff ( float *arrayA, float *arrayB, unsigned long n );

/** Calculate the sum of all absolute differences between all corresponding entries
 * This is calculated on the CPU in parallel
 * \f$ \sum_{i=0}^{n-1} {|A_i-B_i|}  \f$
 * 
 * @param arrayA The first array
 * @param arrayB The second array
 * @param n The number of elements to compare
 */
int absDiff ( int *arrayA, int *arrayB, unsigned long n );

/** Calculate the maximum absolute differences between ant two corresponding entries
 * This is calculated on the CPU in parallel
 * 
 * @param arrayA The first array
 * @param arrayB The second array
 * @param n The number of elements to compare
 */
int maxAbsDiff ( int *arrayA, int *arrayB, unsigned long n );

/** Normalise the array, in parallel over CPU
 * 
 * @param arr the array of float values
 * @param n   the number of elements in the array
 * @return the number of threads that were active at the time of this call
 **/
void normalize ( float *array, unsigned long n );

/** Normalise the array, in parallel over CPU, only considering the central per values to calculate stats.
 *
 *  This allows ignoring outliers, it orders the data and then only considers the central per% values.
 * 
 *  Note the sort can be expensive it is O\f$Nlog_2(N)\f$
 * 
 *  The ordering requires editing the array. If no tmpArray is specified a copy is made, <b>this may be memory and time inefficient!</b>
 * 
 * @param arr the array of float values
 * @param n   the number of elements in the array
 * @param per only use the central per items
 * @return the number of threads that were active at the time of this call
 **/
void normalizePer ( float *array, unsigned long n, float per = 0.999, float *tmpArry = NULL);

#endif // ARRAYOPS_H_INCLUDED

// kate: indent-mode cstyle; indent-width 2; replace-tabs on; 
