#ifndef LOG_H_INCLUDED
#define LOG_H_INCLUDED

#include <iostream>
#include <fstream>
#include <pthread.h>
#include <unistd.h>

#include "util.h"

#if _WIN32
#include <Windows.h>
#else
#include <sys/time.h>
#endif


#define CSVMAXLINE 2048

#define LOG 1

#define SPINEND_NONE      0
#define SPINEND_DONE      1
#define SPINEND_DONENEW   2

void printTextLine ( int lineWidth, char* text, char edge, char fill );
void printTextLine2 ( int before, int after, int lineWidth, char edge, char fill, const char* text, ... );

/** A struct holding info on a spin message
 *
 */
struct spiner
{
    pthread_t thread;
    char      msg[1024];
};

/** A object to write log files
 * This can write with line numbers, time stamp and CSV tables.
 */
class Logger
{
public:
    /** Constructor */
    Logger ( );

    /** Constructor.
     * Create a logger. If a file name exists additional text will be written to the end, if not it will be created.
     * @param filename  The name of the log file to open
     */
    Logger ( char* filename , bool append = false );

    /** Destructor
     * This will close the log file if it is still open.
     */
    ~Logger();

    /** Close log file;
     * This will close the log file if it is still open.
     */
    void close();
    
    
    /** Write text to the log file
    * Write text to the log file, if the file is open.
    * @param text  The text to be written to the log file
    */
    void write ( char* text );

    /** Write text to the log file
     * Write text to the log file, if the file is open.
     * @param format C string that contains a format string that follows the same specifications as format in <a href="http://www.cplusplus.com/printf">printf</a>
     */
    void write ( const char* format, ... );

    /** Write an entry precoded by a time stamp
    * Writes the text preceded by a time stamp of the format:
    * 'YYYY/MM/DD HH:MM:SS | '
    * @param text  The text to be written to the log file
    */
    void TimeWrite ( char* text );

    /** Write an entry proceeded by the line number
    * Writes the text preceded by the line number.
    * This will line must be ended by a newline, it will no be added.
    * Additional new lines in the text will not counted as lines
    * @param line  The line to be written to the log file
    */
    void lineWrite ( char* line );

    /** Write an entry proceeded by the line number
    * Writes the text preceded by the line number.
    * This will line must be ended by a newline, it will no be added.
    * Additional new lines in the text will not counted as lines
    * @param line C string that contains a format string that follows the same specifications as format in <a href="http://www.cplusplus.com/printf">printf</a>
    */
    void lineWrite ( const char* line, ... );

    /** Log a variable's value
    * This is used to create a table in a CSV file.
    * Each call will enter a value in the current row. A row is ended with csvEndLine()
    * The first complete row written will establish the headings, which will be written as a preceding row.
    * If the current row is not the first and a call has a different heading to the previous row, a new table will be created by writing the new list of headings above the current line.
    * @param title  The heading of the csv column ( usually the variable name, or descriptor)
    * @param value C string that contains a format string that follows the same specifications as format in <a href="http://www.cplusplus.com/printf">printf</a>
    */
    void csvWrite ( const char* title, const char* value, ... );

    /** Log a variable's value
    * This is used to create a table in a CSV file.
    * Each call will enter a value in the current row. A row is ended with csvEndLine()
    * The first complete row written will establish the headings, which will be written as a preceding row.
    * If the current row is not the first and a call has a different heading to the previous row, a new table will be created by writing the new list of headings above the current line.
    * This function allows two headings, title1 will be in the firs row while title2 will be in the second. This allows for instance the use of units.
    * @param title1  The first  heading of the csv column ( usually the variable name, or descriptor )
    * @param title2  The second heading of the csv column ( usually the units )
    * @param value C string that contains a format string that follows the same specifications as format in <a href="http://www.cplusplus.com/printf">printf</a>
    */
    void csvWrite ( const char* title1, const char* title2, const char* value, ... );

    /** End the current row of the CSV table
     */
    void csvEndLine();

    /** Write a blank line to the CSV file
     * This writes a blank line, while maintaining the current headings.
     */
    void csvBlankLine();

    /** Get the current line number
    * @param deliminator the character to use in as the delimiter
    * @return The next line number to be written
    */
    unsigned int getLineNo() { return lineNo; };

    /** Set the CSV Delimiter
    * Set the CSV delimiter to use in a CSV file
    * @param delimiter the character to use in as the delimiter
    */
    void sedCsvDeliminator ( char deliminator ) {
        delim = deliminator;
    };

    /** Set the file to be written to
    * Set the file to be written to, if file name exists additional text will be written to the end, if not it will be created.
    * @param filename The name of the file to be opened
    */
    void setFile ( char* filename, bool append = false );

private:
    FILE * myfile;                    ///< The file being written to
    unsigned int lineNo;              ///< The line number used in lineWrite()
    int cvVarNo;                      ///< The current column of the CSV file
    char delim;                       ///< The delimiter for a CSV file, the default is ';'

    char fname[1024];        ///< The buffer to hold the heading

    char headLine1[CSVMAXLINE];        ///< The buffer to hold the heading
    char* tPointer;                   ///< A pointer to point to the next position in the headLine1 to be write
    char headLine2[CSVMAXLINE];       ///< The buffer to hold the heading
    char* tPointer2;                  ///< A pointer to point to the next position in the headLine2 to be write
    char valLine[CSVMAXLINE];         ///< The buffer to hold the values
    char* vPointer;                   ///< A pointer to point to the next position in the valLine to be write

    bool newList;                     ///< A flag to indicate that this is the first row of a CSV table

    /** Initialise variables
    * Initialize all variables.
    */
    void init();
};

spiner* spinMessage( const char* text, ... );
void killSpinMessage(spiner* sp, int end = SPINEND_DONENEW);

#endif // LOG_H_INCLUDED
